#!/bin/bash

set -e -u -f

declare -r terminal_timeout=1.2

# DECKPAM/DECKPNM (KeyPad Application/Numeric Mode)
# '\033='  keypad application mode
# '\033>'  keypad numeric mode
#

pipe_escaped() {
    cat -e | sed -e 's/[$]$//'
    echo
}


print_escaped() {
    local str=$(printf "$@")
    printf '%s' "${str}" | pipe_escaped
}


print_verbose() {
    local str=$(printf "$@")
    print_escaped "$0: %s" "${str}" 1>&${verbose_fd}
}


#quote_shell_words() {
#    sed -e "s/'/'\\\\''/g" \
#        -e 's/\([^ ]\+\)\( \+\|$\)/'\''\1'\'' /g'
#}

send_string() {
    local str=$(printf "$@")
    print_escaped "$0: sending to terminal: ${str}" 1>&${verbose_fd}
    printf '%s' "${str}" > /dev/tty
}


recv_ansi_string() {
    local str
    local char

    str=''
    while true; do
        if ! read -s -r -t "${terminal_timeout}" -N 1 char < /dev/tty; then
            if [ -z "${str}" ]; then
		print_verbose "timeout waiting for ANSI response"
            else
		print_verbose "timeout reading incomplete ANSI response: ${str}"
            fi
            return 1
        fi
        str+="${char}"
        if echo "${char}" | egrep '[a-zA-Z]' > /dev/null; then
            break
        fi
    done

    print_escaped "$0: received ANSI string from terminal: ${str}" 1>&${verbose_fd}
    printf '%s' "${str}"
    while read -s -r -t ".1" -N 1 char < /dev/tty; do
      print_verbose "got extra char after ANSI string: ${str}"
    done
    return 0
}


recv_dcs_string() {
    local str
    local char

    str=''
    while true; do
        if ! read -s -r -t "${terminal_timeout}" -N 1 char < /dev/tty; then
            if [ -z "${str}" ]; then
		print_verbose "timeout waiting for DSC response"
            else
		print_verbose "timeout reading incomplete DSC response: ${str}"
            fi
            return 1
        fi
        str+="${char}"
        if echo "${str}" | egrep $'\033''\\$' > /dev/null; then
            break
        fi
    done

    print_escaped "$0: received DSC string from terminal: ${str}" 1>&${verbose_fd}
    printf '%s' "${str}"
    return 0
}


parse_ansi_string() {
    sed -e 's/^'$'\033''\[//' -e 's/^\(.*\)[a-zA-Z]$/\1/' | tr -d $'\033'
}


parse_dcs_string() {
    sed -e 's/^'$'\033'']//' -e 's/^[A-Za-z][?]*\(.*\)'$'\033''\\$/\1/' | tr -d $'\033'
}


set_memory_lines_per_page() {
    # DECSLPP
    # 24  24 lpp / 6 pages
    # 25  25 lpp / 5 pages
    send_string '\033[%dt' "$1"
    recv_ansi_string || return 1
}


get_dimensions() {
    # dtterm
    #send_string '\033[18t'

    # should be more portable
    #send_string '\033[999;999H\033[6n'
    #send_string '\033[r\033[999;999H\033[6n'
    send_string '\0337\033[r\033[999;999H\033[6n\0338'
    #output:^[[X;YR
    recv_ansi_string || return 1
}


get_window_size() {
    # xterm
    send_string '\033[14t'
    recv_ansi_string || return 1
}


get_window_position() {
    # xterm
    send_string '\033[13t'
    recv_ansi_string || return 1
}


get_window_state() {
    # xterm
    send_string '\033[11t'
    recv_ansi_string || return 1
}


get_window_title() {
    # dtterm
    send_string '\033[21t'
    recv_dcs_string || return 1
}


set_window_title() {
    # xterm
    send_string '\033]0;%s\007' "$1"
}


save_window_title() {
    # xterm
    send_string '\033[22;0t'
}


restore_window_title() {
    # xterm
    send_string '\033[23;0t'
}


get_window_icon_title() {
    # dtterm
    send_string '\033[20t'
    recv_dcs_string || return 1
}


set_window_icon_title() {
    # xterm
    send_string '\033[20t' "$1"
}


save_icontitle() {
    # xterm
    send_string '\033[22;1t'
}


restore_icontitle() {
    # xterm
    send_string '\033[23;1t'
}


get_primary_attributes() {
    # DA (Device Attributes) ->
    # ->
    # architecture:
    #   1 VT100,VT101
    #   6 VT102
    #  12 VT125
    #  62 VT2xx
    #  63 VT3xx
    #  64 VT4xx
    #  65 VT5xx
    #
    # other attributes:
    #  0  no options
    #  1  132-columns
    #  2  printer port
    #  3  ReGIS graphics
    #  4  sixel graphics
    #  6  selective-erase
    #  7  soft character sets
    #  8  user-defined keys
    #  9  national replacement character sets
    # 12  Serbo-croatian (SCS) / Yugoslavian
    # 13  local editing mode
    # 14  eight bit
    # 15  technical character set
    # 16  mouse (locator device)
    # 17  terminal state reports
    # 18  windowing
    # 19  dual sessions
    # 21  horizontal scrolling
    # 22  ANSI color
    # 23  Greek
    # 24  Turkish
    # 29  ANSI text locator
    # 39  page memory extension
    # 42  Latin-2
    # 44  PCTerm
    # 45  soft key map
    # 46  ASCII terminal emulation
    send_string '\033[0c'
    recv_ansi_string || return 1
}

get_secondary_attributes() {
    # DA2 (Device Attributes 2) ->
    # ->
    # device:
    #   1 VT220
    #  18 VT330
    #  19 VT340
    #  24 kermit  (from vttest)
    #  28 DECTerm (from vttest)
    #  41 VT420
    #  64 VT520
    #  65 VT525
    # ROM version
    send_string '\033[>0c'
    recv_ansi_string || return 1
}

# DA3
# DECRPTUI

# DECSCL

set_compatibility_level() {
    # DECSCL (Set Compatibility Level)
    # 61   VT100 mode
    # 62   VT220 mode (8 bit controls)
    # 62;0 VT220 mode (8 bit controls)
    # 62;1 VT220 mode (7 bit controls)
    # 62;2 VT220 mode (8 bit controls)
    send_string '\033[%d;%d"p' "$1" "$2"
}

# SCS

set_terminal_id() {
    # DECTID (select Terminal ID)
    #  0  VT100
    #  1  VT101
    #  2  VT102
    #  5  VT220
    #  7  VT320
    #  9  VT420
    # 10  VT510
    send_string '\033[%d,q' "$1"
}


# DESSCPP (Set Columns Per Page)
# '\033[$|'     80 columns
# '\033[0$|'    80 columns
# '\033[80$|'   80 columns
# '\033[132$|' 132 columns

# NP (Next Page)
# '\033[0U'

# PP (Previous Page)
# '\033[0V'

# DECSASD (Select Active Status Display)
# '\0330$}'  host-writable status line
# '\0331$}'  terminal parameter indicator in status line

# DECSSDT (Set Status Display Type)
# '\0330$~'  blank status line
# '\0331$~'  terminal parameter indicator in status line (default)
# '\0332$~'  host-writable status line


reset_terminal() {
    # DECSTR
    send_string '\033[!p'
}


# DEC Private Mode Settings
#
# DECCKM (Cursor Key Mode)
# '\033[?1h'  cursor keys send application controls
# '\033[?1l'  cursor keys send ANSI cursor controls
#
# DECANM (ANsi/vt52 Mode)
# (there is no way to leave VT52 mode except terminal reset)
# '\033[?2l'  leave ANSI mode and enter VT52 mode
#
# DECCOLM (set COLumn Mode)
# '\033[?3h'  132 columns
# '\033[?3l'   80 columns
#
# DECSCLM (SCroLling Mode)
# '\033[?4h'  smooth scroll (6 lines / s)
# '\033[?4l'  jump scroll
#
# DECSCNM (SCreeN Mode)
# '\033[?5h'  reverse video
# '\033[?5l'  normal video
#
# DECOM (Origin Mode)
# '\033[?6h'  top of scrolling region
# '\033[?6l'  top of screen
#
# DECAWM (Auto Wrap Mode)
# '\033[?7h'  enable auto wrap
# '\033[?7l'  disable auto wrap
#
# DECARM (Auto Repeat Mode)
# '\033[?8h'  enable auto repeat
# '\033[?8l'  disable auto repeat
#
# DECPFF (Print Form Feed)
# '\033[?18h'  send form feed at end of print screen
# '\033[?18l'  do not send form feed
# 
# DECPEX (Print EXtent)
# '\033[?19h'  print screen prints full screen
# '\033[?19l'  print screen prints scrolling region
#
# DECTCEM (Text Cursor Enable Mode)
# '\033[?25h'  cursor visible
# '\033[?25l'  cursor invisible
#
# DECSDM (Sixel Display Mode)
# '\033[?80h'  inline scrolled
# '\033[?80l'  absolute positioning
#
# 42 DECNRCM  NRC set
# 66 DECNKM   numeric keypad
# 67 DECBKM   backarrow key
# 68 DECKBUM  keyboard usage
get_mode_setting() {
    # DECRQM -> DECRPM
    #
    # -> DECRPM
    # '\033[M;V$y'
    #  0 unknown
    #  1 set
    #  2 reset
    #  3 set (permanent)
    #  4 reset (permanent)
    send_string '\033[%s$p' "$1"
    recv_ansi_string || return 1
}


set_ansi_modes() {
    # SM
    local modelist
    local mode

    [ $# -gt 1 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[%sh' "${modelist}"
}


reset_ansi_modes() {
    # RM
    local modelist
    local mode

    [ $# -gt 0 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[%sl' "${modelist}"
}


set_dec_modes() {
    # DECSET
    local modelist
    local mode

    [ $# -gt 0 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[?%sh' "${modelist}"
}


reset_dec_modes() {
    # DECRST
    local modelist
    local mode

    [ $# -gt 0 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[?%sl' "${modelist}"
}


set_xterm_modes() {
    # DECSET
    local modelist
    local mode

    [ $# -gt 1 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[?%sh' "${modelist}"
}


reset_xterm_modes() {
    # DECRST
    local modelist
    local mode

    [ $# -gt 1 ] || return 1
    modelist="$1"
    for mode in "${@:2}"; do
        modelist+=";${mode}"
    done
    send_string '\033[?%sl' "${modelist}"
}


get_graphics() {
    # MC (Graphics to Host)
    send_string '\033[?%di' 2
    recv_dcs_string || return 1
}


send_graphics_to_printer() {
    # MC (Graphics to Printer)
    send_string '\033[?%di' 0
}


get_terminal_status() {
    # DSR (terminal DSR request)
    send_string '\033[%dn' 5
    recv_ansi_string || return 1
}


get_cursor_status() {
    # DSR (cursor CPR request)
    send_string '\033[%dn' 6
    recv_ansi_string || return 1
}


get_printer_status() {
    # DSR (printer status request)
    send_string '\033[?%dn' 15
    recv_ansi_string || return 1
}


get_udk_status() {
    # DSR (user defined keys status request)
    send_string '\033[?%dn' 25
    recv_ansi_string || return 1
}


# response ?27 + dialect + keyboard status(0=ready,3=none,8=busy)
get_keyboard_language() {
    # DSR (printer port status request)
    send_string '\033[?%dn' 26
    recv_ansi_string || return 1
}


get_locator_status() {
    # DSR (locator status request)
    send_string '\033[?%dn' 55
    recv_ansi_string || return 1
}


get_locator_type() {
    # DSR (locator type request)
    send_string '\033[?%dn' 56
    recv_ansi_string || return 1
}


get_macro_checksum() {
    # DSR (macro definition checksum request)
    send_string '\033[?%dn' 63
    recv_dcs_string || return 1
}


extract_window_position() {
    local -r response="$(get_window_position | parse_ansi_string)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -ne 3 ]; then
        print_verbose "expected parameter count 3, got ${#values[@]}"
        return 1
    fi
    echo "${values[1]};${values[2]}"
}


extract_window_size() {
    local -r response="$(get_window_size | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 3 ]; then
        print_verbose "expected parameter count 3, got ${#values[@]}"
        return 1
    fi
    # This is the size without the internal border.
    echo "${values[2]};${values[1]}"
}


extract_window_state() {
    local -r response="$(get_window_state | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi
    local state
    case ${values[0]} in
    (1) state="open";;
    (2) state="iconified";;
    (*) state="unknown";;
    esac
    echo "${state}"
}


extract_window_title() {
    local -r response="$(get_window_title | parse_dcs_string)"
    local -a values
    IFS=' ' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi
    echo "${values[0]}"
}


extract_window_icon_title() {
    local -r response="$(get_window_icon_title | parse_dcs_string)"
    local -a values
    IFS=' ' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi
    echo "${values[0]}"
}


extract_dimensions() {
    local -r response="$(get_dimensions | parse_ansi_string)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -ne 2 ]; then
        print_verbose "expected parameter count 2, got ${#values[@]}"
        return 1
    fi
    echo "${values[0]};${values[1]}"
}


extract_primary_attributes() {
    local -r response="$(get_primary_attributes | parse_ansi_string)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -lt 1 ]; then
        print_verbose "no parameters returned"
        return 1
    fi

    local arch
    local argmode
    case ${values[0]} in
    ( ?1) arch=VT100 argmode=VT100;;
    ( ?2) arch=VT102 argmode=unknown;;  # always has AVO
    ( ?4) arch=VT132 argmode=VT100;;  # always has AVO
    ( ?5) arch=VK100 argmode=unknown;;
    ( ?6) arch=VT102 argmode=none;;  # always has AVO
    ( ?7) arch=VT131 argmode=none;;  # always has AVO
    (?12) arch=VT125 argmode=VT125;;
    (?61) arch=VT1xx argmode=op_level;;
    (?62) arch=VT2xx argmode=op_level;;
    (?63) arch=VT3xx argmode=op_level;;
    (?64) arch=VT4xx argmode=op_level;;
    (?65) arch=VT5xx argmode=op_level;;
    (*) arch="\"${values[0]}\"" argmode=unknown; print_verbose "unknown arch: ${values[0]}";;
    esac

    local rom
    local -a options
    local attr
    case ${argmode} in
    (none)
        ;;
    (VT100)
	if ((${values[1]} & 1)); then
	    options+=('STP (processor options)')
        fi
	if ((${values[1]} & 2)); then
	    options+=('AVO (advanced video option)')
        fi
	if ((${values[1]} & 4)); then
	    options+=('GPO (ReGIS graphics processor option)')
        fi
	if ((${values[1]} & 8)); then
	    options+=('PP (printer port)')
        fi

        rom=""
        ;;
    (VT125)
	if ((${values[1]} & 1)); then
	    options+=('STP (processor option)')
        fi
	if ((${values[1]} & 2)); then
	    options+=('AVO (advanced video option)')
        fi
	if ((${values[1]} & 4)); then
	    options+=('GPO (graphics processor option)')
        fi
	if ((${values[1]} & 8)); then
	    options+=('PP (printer port)')
        fi

	if [ X"${values[2]}" = X"0" ]; then
	    options+=('no printer attached')
        else
	    options+=('printer attached')
        fi

	rom="${values[3]}"
        ;;
    (op_level)
	for attr in "${values[@]:2}"; do
	    case ${attr} in
	    ( 0) options+=('no options');;
	    ( 1) options+=('132-columns');;
	    ( 2) options+=('printer port');;
	    ( 3) options+=('ReGIS graphics');;
	    ( 4) options+=('sixel graphics');;
	    ( 6) options+=('selective-erase');;
	    ( 7) options+=('soft character sets');;
	    ( 8) options+=('user-defined keys');;
	    ( 9) options+=('national replacement character sets');;
	    (12) options+=('Serbo-croatian (SCS) / Yugoslavian');;
	    (13) options+=('local editing mode');;
	    (14) options+=('eight bit');;
	    (15) options+=('technical character set');;
	    (16) options+=('mouse (locator device)');;
	    (17) options+=('terminal state reports');;
	    (18) options+=('windowing');;
	    (19) options+=('dual sessions');;
	    (21) options+=('horizontal scrolling');;
	    (22) options+=('ANSI color');;
	    (23) options+=('Greek');;
	    (24) options+=('Turkish');;
	    (29) options+=('ANSI text locator');;
	    (39) options+=('page memory extension');;
	    (42) options+=('Latin-2');;
	    (44) options+=('PCTerm');;
	    (45) options+=('soft key map');;
	    (46) options+=('ASCII terminal emulation');;
	    (*) options+=("attribute \"${attr}\"");;
	    esac
	done
        rom=""
        ;;
    (unknown)
	print_verbose "unable to parse attributes: ${values[@]:2}"
        ;;
    esac

    echo -n "${arch};${rom}"
    printf ';%s' "${options[@]}"
}


extract_secondary_attributes() {
    local -r response="$(get_secondary_attributes | parse_ansi_string)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -lt 1 ]; then
        print_verbose "no parameters returned"
        return 1
    fi

    local dev
    case ${values[0]} in
    ( ?0) dev="xterm (VT100 mode)";;
    ( ?1) dev='VT220';;
    ( ?2) dev='VT240';;
    (?18) dev='VT330';;
    (?19) dev='VT340';;
    (?24) dev='VT320';;
    (?41) dev='VT420';;
    (?64) dev='VT520';;
    (?65) dev='VT525';;
    (*) dev="\"${values[0]}\"";;
    esac
    local rom="${values[1]}"
    echo "${dev};${rom}"
}


extract_mode_setting() {
    local -r mode=$1
    local -r response="$(get_mode_setting "${mode}" | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -lt 1 ]; then
	print_verbose "no values in mode ${mode} status: ${response}"
	return
    fi
    if [ ${#values[@]} -gt 2 ]; then
	print_verbose "unexpected extra parameters in mode ${mode} status: ${response}"
    fi

    local name
    local desc
    local state
    case ${values[0]} in
    ( 1) name="ANSI mode ${values[0]}"; desc='guarded area transfer';;
    ( 2) name="ANSI mode ${values[0]}"; desc='keyboard action';;
    ( 3) name="ANSI mode ${values[0]}"; desc='control representation';;
    ( 4) name="ANSI mode ${values[0]}"; desc='insert/replace';;
    ( 5) name="ANSI mode ${values[0]}"; desc='status reporting transfer';;
    ( 6) name="ANSI mode ${values[0]}"; desc='erasure';;
    ( 7) name="ANSI mode ${values[0]}"; desc='vertical editing';;
    ( 8) name="ANSI mode ${values[0]}"; desc='bidirectional support';;
    ( 9) name="ANSI mode ${values[0]}"; desc='unknown';;  # FIXME
    (10) name="ANSI mode ${values[0]}"; desc='horizontal editing';;
    (11) name="ANSI mode ${values[0]}"; desc='positioning unit';;
    (12) name="ANSI mode ${values[0]}"; desc='send/receive';;
    (13) name="ANSI mode ${values[0]}"; desc='format effector action';;
    (14) name="ANSI mode ${values[0]}"; desc='format effector transfer';;
    (15) name="ANSI mode ${values[0]}"; desc='multiple area transfer';;
    (16) name="ANSI mode ${values[0]}"; desc='transfer termination';;
    (17) name="ANSI mode ${values[0]}"; desc='selected area transfer';;
    (18) name="ANSI mode ${values[0]}"; desc='tabulation stop';;  # xterm: DECPFF (Don't print form feed)
    (19) name="ANSI mode ${values[0]}"; desc='editing boundary';;  # xterm: DECPEX (Limit print to scrolling region)
    (20) name="ANSI mode ${values[0]}"; desc='line feed/newline';;
    (21) name="ANSI mode ${values[0]}"; desc='graphic rendition combination';;

    (  \?1) name=DECCKM; desc='cursor keys';;
    (  \?2) name=DECANM; desc='ANSI';;
    (  \?3) name=DECCOLM; desc='column';;
    (  \?4) name=DECSCLM; desc='smooth scrolling';;
    (  \?5) name=DECSCNM; desc='screen';;
    (  \?6) name=DECOM; desc='origin';;
    (  \?7) name=DECAWM; desc='autowrap';;
    (  \?8) name=DECARM; desc='autorepeat';;
    (  \?9) name="DEC mode ${values[0]#\?}"; desc='interlace';;  # xterm says send mouse X&Y
    ( \?10) name="DEC mode ${values[0]#\?}"; desc='editing';;  # VT131, rxvt says "show toolbar"
    ( \?11) name="DEC mode ${values[0]#\?}"; desc='line transmit';;  # VT131
    ( \?12) name="DEC mode ${values[0]#\?}"; desc='start blinking cursor';;  # ATT610
    ( \?13) name="DEC mode ${values[0]#\?}"; desc='space compression/field delimiter';;  # VT131
    ( \?14) name="DEC mode ${values[0]#\?}"; desc='transmit execution';;  # VT131
    ( \?16) name="DEC mode ${values[0]#\?}"; desc='edit key execution';;  # VT131
    ( \?18) name=DECPFF; desc='print form feed';;
    ( \?19) name=DECPEX; desc='printer extent';;
    ( \?24) name="DEC mode ${values[0]#\?}"; desc='printer density';;  # LA12
    ( \?25) name=DECTCEM; desc='text cursor enable';;
    ( \?27) name="DEC mode ${values[0]#\?}"; desc='proportional spacing';;  # LQP02
    ( \?29) name="DEC mode ${values[0]#\?}"; desc='pitch select';;
    ( \?30) name="rxvt mode ${values[0]#\?}"; desc='scroll bar';;  # rxvt  FIXME: is there an actual DEC mode?
    ( \?34) name="DEC mode ${values[0]#\?}"; desc='cursor direction, right to left';;
    ( \?35) name="DEC mode ${values[0]#\?}"; desc='Hebrew keyboard mapping';;  # rxvt says "enable font-shifting functions"
    ( \?36) name="DEC mode ${values[0]#\?}"; desc='Hebrew encoding mode';;
    ( \?38) name=DECTEK; desc='Tektronix';;
    ( \?40) name="DEC mode ${values[0]#\?}"; desc='carriage return/newline';;  # xterm says "allow 80 -> 132 mode"
    ( \?41) name="DEC mode ${values[0]#\?}"; desc='print direction';;  # xterm says "more(1) fix"
    ( \?42) name=DECN-RCM; desc='national replacement character sets';;
    ( \?43) name="DEC mode ${values[0]#\?}"; desc='expanded print mode';;
    ( \?44) name="DEC mode ${values[0]#\?}"; desc='color print mode';;  # xterm says "turn on margin bell"
    ( \?45) name="DEC mode ${values[0]#\?}"; desc='print colorspace';;  # xterm says "reverse wraparound"
    ( \?46) name="DEC mode ${values[0]#\?}"; desc='print background';;  # xterm says "start logging"
    ( \?47) name="DEC mode ${values[0]#\?}"; desc='print rotated';;  # xterm says "use alternate screen buffer"
    ( \?51) name="DEC mode ${values[0]#\?}"; desc='black/white swap';;
    ( \?52) name="DEC mode ${values[0]#\?}"; desc='origin placement';;
    ( \?55) name="DEC mode ${values[0]#\?}"; desc='bold page';;
    ( \?57) name="DEC mode ${values[0]#\?}"; desc='Greek keyboard mapping';;
    ( \?60) name="DEC mode ${values[0]#\?}"; desc='horizontal cursor coupling';;
    ( \?61) name="DEC mode ${values[0]#\?}"; desc='vertical cursor coupling';;
    ( \?64) name="DEC mode ${values[0]#\?}"; desc='page cursor coupling';;
    ( \?66) name=DECNKM; desc='application keypad';;
    ( \?67) name=DECBKM; desc='backarrow key sends backspace';;
    ( \?68) name="DEC mode ${values[0]#\?}"; desc='process keyboard data';;
    ( \?69) name=DECLRMM; desc='vertical split screen';;
    ( \?73) name="DEC mode ${values[0]#\?}"; desc='transmit rate limiting';;
    ( \?80) name="DEC mode ${values[0]#\?}"; desc='sixel scrolling';;
    ( \?81) name="DEC mode ${values[0]#\?}"; desc='key position';;
    ( \?95) name=DECNCSM; desc='no clearing screen on column change';;
    ( \?96) name="DEC mode ${values[0]#\?}"; desc='cursor right to left';;
    ( \?97) name="DEC mode ${values[0]#\?}"; desc='CRT save';;
    ( \?98) name="DEC mode ${values[0]#\?}"; desc='auto resize';;
    ( \?99) name="DEC mode ${values[0]#\?}"; desc='modem control';;
    (\?100) name="DEC mode ${values[0]#\?}"; desc='auto answerback';;
    (\?101) name="DEC mode ${values[0]#\?}"; desc='conceal answerback message';;
    (\?102) name="DEC mode ${values[0]#\?}"; desc='ignoring null';;
    (\?103) name="DEC mode ${values[0]#\?}"; desc='half-duplex';;
    (\?104) name="DEC mode ${values[0]#\?}"; desc='secondary keyboard language';;
    (\?106) name="DEC mode ${values[0]#\?}"; desc='overscan';;

    (\?1000) name="xterm mode ${values[0]#\?}"; desc='X-Y mouse click reporting';;
    (\?1001) name="xterm mode ${values[0]#\?}"; desc='Use Hilite Mouse Tracking';;
    (\?1002) name="xterm mode ${values[0]#\?}"; desc='Use Cell Motion Mouse Tracking';;
    (\?1003) name="xterm mode ${values[0]#\?}"; desc='Use All Motion Mouse Tracking';;
    (\?1004) name="xterm mode ${values[0]#\?}"; desc='Send FocusIn/FocusOut events';;
    (\?1005) name="xterm mode ${values[0]#\?}"; desc='Enable UTF-8 Mouse Mode';;
    (\?1006) name="xterm mode ${values[0]#\?}"; desc='Enable SGR Mouse Mode';;
    (\?1007) name="xterm mode ${values[0]#\?}"; desc='Enable Alternate Scroll Mode';;
    (\?1010) name="rxvt mode ${values[0]#\?}"; desc='Scroll to bottom on tty output';;  # rxvt
    (\?1011) name="rxvt mode ${values[0]#\?}"; desc='Scroll to bottom on key press';;  # rxvt
    (\?1015) name="urxvt mode ${values[0]#\?}"; desc='Enable urxvt Mouse Mode';;  # urxvt
    (\?1034) name="xterm mode ${values[0]#\?}"; desc='eightBitInput';;
    (\?1035) name="xterm mode ${values[0]#\?}"; desc='numLock';;
    (\?1036) name="xterm mode ${values[0]#\?}"; desc='metaSendsEscape';;
    (\?1037) name="xterm mode ${values[0]#\?}"; desc='Send DEL from the editing-keypad Delete key';;
    (\?1039) name="xterm mode ${values[0]#\?}"; desc='altSendsEscape';;
    (\?1040) name="xterm mode ${values[0]#\?}"; desc='keepSelection';;
    (\?1041) name="xterm mode ${values[0]#\?}"; desc='selectToClipboard';;
    (\?1042) name="xterm mode ${values[0]#\?}"; desc='bellIsUrgent';;
    (\?1043) name="xterm mode ${values[0]#\?}"; desc='popOnBell';;
    (\?1047) name="xterm mode ${values[0]#\?}"; desc='Use alternate screen buffer. (This may be disabled by the titeInhibit resource)';;
    (\?1048) name="xterm mode ${values[0]#\?}"; desc='Save cursor as in DECSC. (This may be disabled by the titeInhibit resource)';;
    (\?1049) name="xterm mode ${values[0]#\?}"; desc='Save cursor as in DECSC and use alternate screen buffer, clearing it first. (This may be disabled by the titeInhibit resource)';;
    (\?1050) name="xterm mode ${values[0]#\?}"; desc='terminfo/termcap function-key mode';;
    (\?1051) name="xterm mode ${values[0]#\?}"; desc='Sun function-key mode';;
    (\?1052) name="xterm mode ${values[0]#\?}"; desc='HP function-key mode';;
    (\?1053) name="xterm mode ${values[0]#\?}"; desc='SCO function-key mode';;
    (\?1060) name="xterm mode ${values[0]#\?}"; desc='legacy keyboard emulation (X11R6)';;
    (\?1061) name="xterm mode ${values[0]#\?}"; desc='VT220 keyboard emulation';;
    (\?2004) name="xterm mode ${values[0]#\?}"; desc='bracketed paste mode';;

    (*) name="unknown mode ${values[0]}"; desc="unknown";;
    esac

    if [ ${#values[@]} -gt 1 ]; then
	case ${values[1]} in
	(0\$) state='unrecognized';;
	(1\$) state='set';;
	(2\$) state='reset';;
	(3\$) state='set permanently';;
	(4\$) state='reset permanently';;

	(*) state="state \"${values[1]}\"";;
	esac
    else
      state='none'
    fi

    echo "${name};${state};${desc}"
}


extract_terminal_status() {
    local -r response="$(get_terminal_status | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi

    local status
    case ${values[0]} in
    (0) status='ok';;
    (3) status='malfunction';;
    (*) status="\"${values[0]}\"";;
    esac

    echo "${status}"
}


extract_cursor_status() {
    local -r response="$(get_cursor_status | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 2 ]; then
        print_verbose "expected parameter count 2, got ${#values[@]}"
        return 1
    fi

    echo "${values[0]};${values[1]}"
}


extract_printer_status() {
    local -r response="$(get_printer_status | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi

    local status
    case ${values[0]} in
    (\?10) status='printer ready';;
    (\?11) status='print not ready';;
    (\?13) status='no printer detected';;
    (\?18) status='printer busy';;
    (\?19) status='printer assigned to different session';;
    (*) status="\"${values[0]}\"";;
    esac

    echo "${status}"
}


extract_udk_status() {
    local -r response="$(get_udk_status | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi

    local status
    case ${values[0]} in
    (\?20) status='unlocked';;
    (\?21) status='locked';;
    (*) status="\"${values[0]}\"";;
    esac

    echo "${status}"
}

# Mouse:
# 5x

# DECRQTSR: ?%du[;%d] (0=none;1=term;2=color) (0=HLS;1=HLS;2=RGB)
#  terminal: dcs
#  color: dcs

# DECTSR
#  checksum


extract_keyboard_language() {
    local -r response="$(get_keyboard_language | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -lt 2 -o ${#values[@]} -gt 4 ]; then
        print_verbose "expected parameter count 2, 3, or 4, got ${#values[@]}"
        return 1
    fi
    if [ X"${values[0]}" != X"?27" ]; then
        print_verbose "expected response type ?27, got ${values[0]}"
        return 1
    fi

    local dialect
    case ${values[1]} in
    ( 0) dialect='unknown';;
    ( 1) dialect='North American';;
    ( 2) dialect='British';;
    ( 3) dialect='Flemish';;
    ( 4) dialect='French Canadian';;
    ( 5) dialect='Danish';;
    ( 6) dialect='Finnish';;
    ( 7) dialect='German';;
    ( 8) dialect='Dutch';;
    ( 9) dialect='Italian';;
    (10) dialect='Swiss French';;
    (11) dialect='Swiss German';;
    (12) dialect='Swedish';;
    (13) dialect='Norwegian';;
    (14) dialect='French/Belgian';;
    (15) dialect='Spanish';;
    (16) dialect='Portuguese';;
    (*) dialect="\"${values[1]}\"";;
    esac

    local status
    if [ ${#values[@]} -gt 2 ]; then
      case ${values[2]} in
      (0) status='keyboard ready';;
      (3) status='no keyboard';;
      (8) status='keyboard busy';;
      (*) status="\"${values[2]}\"";;
      esac
    else
      status='unknown'
    fi

    # FIXME: values?
    # CSI ? 2 7 ; 1 ; 0 ; 0 n (North American). 
    # The last two parameters apply to VT400 & up, and denote keyboard ready and LK01 respectively. 
    local lktype
    if [ ${#values[@]} -gt 3 ]; then
      case ${values[2]} in
      (-) lktype="LK201";;
      (*) lktype="\"${values[3]}\"";;
      esac
    else
      lktype='unknown'
    fi

    echo "${dialect};${status};${lktype}"
}


extract_locator_status() {
    local -r response="$(get_locator_status | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 1 ]; then
        print_verbose "expected parameter count 1, got ${#values[@]}"
        return 1
    fi

    local status
    case ${values[0]} in
    (\?50) status='ready';;
    (\?53) status='no locator';;
    (\?58) status='busy';;
    (*) status="\"${values[1]}\"";;
    esac

    echo "${status}"
}


extract_locator_type() {
    local -r response="$(get_locator_type | parse_ansi_string)"
    local -a values
    IFS=';' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 2 ]; then
        print_verbose "expected parameter count 2, got ${#values[@]}"
        return 1
    fi
    if [ X"${values[0]}" != X"?57" ]; then
        print_verbose "expected response type ?57, got ${values[0]}"
        return 1
    fi

    local type
    case ${values[1]} in
    (0) type='no locator';;
    (1) type='mouse';;
    (2) type='tablet';;
    (*) type="\"${values[1]}\"";;
    esac

    echo "${type}"
}


extract_macro_checksum() {
    local -r response="$(get_macro_checksum | parse_dcs_string)"
    local -a values
    IFS=' ' read -a values<<<"${response}"
    if [ ${#values[@]} -ne 2 ]; then
        print_verbose "expected parameter count 2, got ${#values[@]}"
        return 1
    fi
    if [ X"${values[0]}" != X"?57" ]; then
        print_verbose "expected response type ?57, got ${values[0]}"
        return 1
    fi

#FIXME
    local type
    case ${values[1]} in
    (0) type='no locator';;
    (1) type='mouse';;
    (*) type="\"${values[1]}\"";;
    esac

    echo "locator_type: ${type}"
}


print_window_position() {
    local result
    local -a values
    local x
    local y

    result=$(extract_window_position)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract window position"
        return 1
    fi
    x=${values[0]}
    y=${values[1]}
    echo "window position: ${x},${y}"
}


print_window_size() {
    local result
    local -a values
    local width
    local height

    result=$(extract_window_size)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract window size"
        return 1
    fi
    width=${values[0]}
    height=${values[1]}
    echo "window size: ${width}x${height}"
}


print_window_state() {
    local state

    state=$(extract_window_state)
    echo "window state: ${state}"
}


print_window_title() {
    local title

    title=$(extract_window_title)
    echo "window title: ${title}"
}


print_window_icon_title() {
    local title

    title=$(extract_window_icon_title)
    echo "window icon title: ${title}"
}


print_dimensions() {
    local result
    local -a values
    local rows
    local cols

    result=$(extract_dimensions)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract dimensions"
        return 1
    fi
    rows=${values[0]}
    cols=${values[1]}
    echo "dimensions: ${cols}x${rows}"
}


print_primary_attributes() {
    local result
    local -a values
    local arch
    local rom
    local option

    result=$(extract_primary_attributes)
    print_verbose "extracted primary string: ${result}"
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract primary device attributes"
        return 1
    fi
    arch=${values[0]}
    rom=${values[1]}

    echo "arch: ${arch}"
    if [ -n "${rom}" ]; then
        echo "ROM version: ${rom}"
    fi
    echo "options:"
    for option in "${values[@]:2}"; do
      echo "  ${option}"
    done
}


print_secondary_attributes() {
    local result
    local -a values
    local device
    local rom

    result=$(extract_secondary_attributes)
    print_verbose "extracted secondary string: ${result}"
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract secondary device attributes"
        return 1
    fi
    device=${values[0]}
    rom=${values[1]}

    echo "device: ${device}"
    echo "ROM version: ${rom}"
}


print_mode_setting() {
    local mode=$1
    local prefix=$2
    local result
    local -a values
    local name
    local state
    local description

    result=$(extract_mode_setting "${mode}")
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 3 ]; then
        print_verbose "unable to extract mode ${mode} setting"
        return 1
    fi
    name=${values[0]}
    state=${values[1]}
    description=${values[2]}
    if [ X"${state}" != X"unrecognized" ]; then
        echo "${prefix}${name}: ${state} (${description})"
    fi
}


print_ansi_mode_settings() {
    echo "ANSI mode settings:"
    for num in {1..21}; do
        print_mode_setting "${num}" '  '
    done
}


print_dec_mode_settings() {
    # http://bjh21.me.uk/all-escapes/all-escapes.txt
    echo "DEC mode settings:"
    for num in {9..11} 13 14 16 18 19 24 25 27 29 34 35 36 38 {40..47} 51 52 55 57 60 61 64 66 67 68 69 73 80 81 95 96 97 98 99 100 101 102 103 104 106; do
        print_mode_setting "?${num}" '  '
    done
}


print_xterm_mode_settings() {
    # http://invisible-island.net/xterm/ctlseqs/ctlseqs.html
    echo "xterm mode settings:"
    for num in 30 1000 1001 1002 1003 1004 1005 1006 1007 1010 1011 1015 1034 1035 1036 1037 1039 1040 1041 1042 1043 1047 1048 1049 1050 1051 1052 1053 1060 1061 2004; do
        print_mode_setting "?${num}" '  '
    done
}


print_terminal_status() {
    local status

    status=$(extract_terminal_status)
    echo "terminal status: ${status}"
}


print_cursor_status() {
    local result
    local -a values
    local row
    local column

    result=$(extract_cursor_status)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract cursor status"
        return 1
    fi
    row=${values[0]}
    column=${values[1]}
    echo "cursor status: ${row},${column}"
}


print_printer_status() {
    local status

    status=$(extract_printer_status)
    echo "printer status: ${status}"
}


print_udk_status() {
    local status

    status=$(extract_udk_status)
    echo "UDK status: ${status}"
}


print_keyboard_language() {
    local result
    local -a values
    local dialect
    local status
    local lktype

    result=$(extract_keyboard_language)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 3 ]; then
        print_verbose "unable to extract keyboard language"
        return 1
    fi
    dialect=${values[0]}
    status=${values[1]}
    lktype=${values[2]}
    echo "keyboard language: ${dialect} (${status}) (${lktype})"
}


print_locator_status() {
    local status

    status=$(extract_locator_status)
    echo "locator status: ${status}"
}


print_locator_type() {
    local type

    type=$(extract_locator_type)
    echo "locator type: ${type}"
}


print_macro_checksum() {
    local checksum

    checksum=$(extract_macro_checksum)
    echo "macro checksum: ${checksum}"
}

