#!/bin/bash

set -e -u -f

declare -r vtdir=$(sed -e 's@[^/]*$@@' <<<"$0")

source ${vtdir}/vtlib.sh


declare display_mode
declare color_mode
declare prompt
declare -a filenames
declare maxframes
declare max_height
declare max_width
declare min_height
declare min_width
declare target_height
declare target_width
declare screen_rows


print_sixel_support() {
    local -r response="$(extract_primary_attributes)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -lt 1 ]; then
        print_verbose "unable to extract primary device attributes"
        return 1
    fi

    local option
    for option in "${values[@]:2}"; do
        if [ X"${option}" = X"sixel graphics" ]; then
            echo "supported"
            return
        fi
    done

    echo "unsupported"
}


print_register_count() {
    local -r response="$(extract_secondary_attributes)"
    local -a values
    IFS=';' read -a values <<<"${response}"
    if [ ${#values[@]} -lt 1 ]; then
        print_verbose "unable to extract secondary device attributes"
        return 1
    fi

    local count
    case ${values[0]} in
    (VT240) count='4';;
    (VT330) count='4';;
    (VT340) count='16';;
    (*) count='0'
    esac
    print_verbose "device ${values[0]}, colors ${count}"
    echo "${count}"
}


clear_screen() {
    send_string '\033[2J'
}


move_to() {
    local -r row=$1
    local -r col=$2

    send_string '\033[%d;%dH' "${row}" "${col}"
}


current_col() {
    local result
    local -a values

    result=$(extract_cursor_status)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract cursor status"
        return 1
    fi
    echo "${values[1]}"
}


current_row() {
    local result
    local -a values

    result=$(extract_cursor_status)
    IFS=';' read -a values <<<"${result}"
    if [ ${#values[@]} -lt 2 ]; then
        print_verbose "unable to extract cursor status"
        return 1
    fi
    echo "${values[0]}"
}


resize_image() {
    if [ X"${display_mode}" = X"fullscreen" ]; then
        pnmscale -nomix -xysize "${target_width}" "${target_height}"
    else
        cat
    fi
}


convert() {
    local -r filename=$1

    print_verbose "changing to image with ${maxcolors} colors"
    output=$(anytopnm < "${filename}" 2>&${verbose_fd} |
             resize_image 2>&${verbose_fd} |
             pnmquant -fs "${maxcolors}" 2>&${verbose_fd} |
             ppmtosixel 2>&${verbose_fd} |
             tr -d $'\220\234')
    if [ -z "${output}" ]; then
        print_verbose "ppm utilities returned no data for ${filename}"
        return 0
    fi

    printf '%s' "${output}"
    return 0
}


save_frames() {
    local -r filename=$1
    local -r outdir=$2
    local width

    if [ "${target_width}" -gt "0" ]; then
        width="${target_width}"
    else
        width="120"
    fi
    mplayer -frames "${maxframes}" \
            -nosound \
            -zoom -xy "${width}" \
            -ao null \
            -vo pnm:outdir="${outdir}" \
            -- "${filename}" > /dev/null 2>&${verbose_fd}
}


show_image() {
    local -r filename=$1
    local pixel_size=$(extract_window_size)
    local maxcolors=$(print_register_count)
    local output
    local orig_col
    local orig_row
    local col
    local char

    if [ "${maxcolors}" -lt 1 ]; then
        maxcolors=64  # assume xterm, though it might not work at all
    fi

    if [ X"${display_mode}" = X"inline" ]; then
	orig_col=$(current_col)
	orig_row=$(current_row)
    fi

    if [ -r "${filename}" ]; then
        output=$(convert "${filename}" 2>/dev/null)

	if [ X"${display_mode}" = X"fullscreen" ]; then
	    clear_screen
	    move_to "1" "1"
        fi
        if [ -z "${output}" ]; then
            echo "$0: unable to parse image file \"${filename}\"" 1>&2
        else
            send_string '\033P%s\033\\' "${output}"
        fi
    else
	if [ X"${display_mode}" = X"fullscreen" ]; then
	    clear_screen
	    move_to "1" "1"
        fi
        echo "$0: unable to read image file \"${filename}\"" 1>&2
    fi

    if [ X"${display_mode}" = X"fullscreen" ]; then
        if [ X"${prompt}" = X"true" ]; then
            read -s -r -N 1 char < /dev/tty || true
            if [ X"${char}" = X"q" ]; then
                break
            fi
        fi
    elif [ X"${display_mode}" = X"fullrows" ]; then
        echo ''
    elif [ X"${display_mode}" = X"inline" ]; then
        col=$(current_col)
        print_verbose "${col} -lt ${orig_col}"
        if [ "${col}" -lt "${orig_col}" ]; then
            printf '\r'
        else
            print_verbose "moving to orig_row=%s curr_col=%s" "${orig_row}" "${col}"
            move_to "${orig_row}" "${col}"
        fi
    fi
}


prepare_tty() {
    stty -echo
    trap cleanup_tty EXIT
}


prepare_terminal() {
    if [ X"${display_mode}" = X"fullscreen" ]; then
        send_string '\0337'  # save cursor location
    fi
    # FIXME: lots of xterm-only escape codes here
    send_string '\033[?25;47;80;1070s'  # save modes (xterm-only)
    send_string '\033[?80h'  # set sixel scrolling mode
    if [ X"${color_mode}" = X"private" ]; then
	send_string '\033[?1070h'  # private color register mode (xterm-only)
    else
	send_string '\033[?1070l'  # shared color register mode (vtxxx default)
    fi
    # FIXME: xterm-only
    if [ X"${display_mode}" = X"fullscreen" ]; then
      send_string '\033[?47h'  # set alternate buffer mode (xterm-only)
      send_string '\033[?25l'  # reset visible cursor mode
    fi

    trap cleanup_terminal EXIT
}


cleanup_terminal() {
    # FIXME: xterm-only
    send_string '\033[?25;47;80;1070r'  # restore modes (xterm-only)
    if [ X"${display_mode}" = X"fullscreen" ]; then
        send_string '\0338'  # restore cursor location
    fi

    cleanup_tty
}

cleanup_tty() {
    stty echo
}


usage() {
    echo "usage: $0 [<options>] [--] [<file> ...]"
    egrep '[#]#' < "$0" |
        sed -e 's/^[^#]*[#]# /  /' |
        awk -F':' '{ printf("%-30s %s\n", $1, $2); }'
}


handle_args() {
    if [ $# -eq 0 ]; then
        usage 1>&2
        exit 0
    fi

    # defaults
    display_mode=fullrows
    color_mode=private
    prompt=true
    maxframes=20
    max_height=0
    max_width=0
    min_height=0
    min_width=0

    while [ $# -gt 0 ]; do
        case $1 in
            (--inline)  ## --inline: use inline display mode
                display_mode=inline
                ;;
            (--fullrows)  ## --fullrows: use full-rows display mode (default)
                display_mode=fullrows
                ;;
            (--fullscreen)  ## --fullscreen: use full-screen display mode
                display_mode=fullscreen
                ;;
            (-h|--help)  ## -h, --help: display this help text and exit
                usage
                exit 0
                ;;
            (--noprompt)  ## --noprompt: don't wait for keypress in fullscreen mode
                prompt=false
                ;;
            (--private-colors)  ## --shared-colors: use private color registers (default)
                color_mode=private
                ;;
            (--shared-colors)  ## --shared-colors: don't use private color registers
                color_mode=shared
                ;;
            (-v|--verbose)  ## -v,--verbose: show processing details
                exec {verbose_fd}>&2
                ;;

            (--inline=*) ;&
            (--fullrows=*) ;&
            (--fullscreen=*) ;&
            (-h=*|--help=*) ;&
            (--noprompt=*) ;&
            (--private-colors=*) ;&
            (--shared-colors=*) ;&
            (-v=*|--verbose=*)
                echo "$0: ${1%=*} does not accept an argument" 1>&2
                return 1
                ;;

            (--maxframes=*)  ## --maxframes=NUM: maximum number of video frames to display
                maxframes="${1#*=}"
                ;;
            # FIXME: implement scaling options
            (--max-height=*)  ## --max-height=NUM: maximum height of image
                max_height="${1#*=}"
                ;;
            (--max-width=*)  ## --max-width=NUM: maximum width of image
                max_width="${1#*=}"
                ;;
            (--min-height=*)  ## --min-height=NUM: minimum height of image
                min_height="${1#*=}"
                ;;
            (--max-width=*)  ## --min-width=NUM: minimum width of image
                min_width="${1#*=}"
                ;;

            (--maxframes) ;&
            (--max-height) ;&
            (--max-width) ;&
            (--min-height) ;&
            (--min-width)
                echo "$0: $1 requires an argument" 1>&2
                return 1
                ;;

            (--)
                shift
                break
                ;;
            (-*)
                echo "$0: unknown option ${1%=*}" 1>&2
                return 1
                ;;
            (*)
                break
                ;;
        esac
        shift
    done

    if [ $# -lt 1 ]; then
        echo "$0: missing filename arguments" 1>&2
        return 1
    fi

    for file in "$@"; do
        filenames+=("${file}")
    done
}


show_files() {
    local pixel_size
    local dimensions
    local file
    local tmpdir

    prepare_tty
    if [ X"$(print_sixel_support)" != X"supported" ]; then
        echo "$0: requires a terminal with sixel graphics support" 1>&2
        exit 2
    fi
    prepare_terminal

    if [ X"${display_mode}" = X"fullscreen" ]; then
        pixel_size=$(extract_window_size)
        target_width=${pixel_size%;*}
        target_height=${pixel_size#*;}
    else
        target_width=0
        target_height=0
    fi
    dimensions=$(extract_dimensions)
    screen_rows=${dimensions%;*}

    for file in "${filenames[@]}"; do
        print_verbose "showing image from ${file}"
        if echo "${file}" | egrep '.(mpg|mpeg)$' > /dev/null; then
            tmpdir=$(mktemp -d)
            save_frames "${file}" "${tmpdir}"
            for frame in $(find "${tmpdir}" -type f); do
                show_image "${frame}"
            done
            #rm -rf "${tmpdir}"
        else
            show_image "${file}"
        fi
    done
}


exec {verbose_fd}> /dev/null

handle_args "$@"
show_files "${filenames[@]}"

exit 0
